
import sys

from epyk.core.data import Data
from epyk.core.js import JsUtils


class DataClass(object):

  def __init__(self, report, attrs=None, options=None):
    self._report, self.options, self._attrs = report, options, attrs or {}
    self.__sub_levels, self.__sub__enum_levels = set(), set()

  def __getitem__(self, i):
    return self._attrs[i]

  def attrs(self):
    return self._attrs.items()

  def custom(self, name, value):
    """
    Description:
    ------------
    Custom function to add a bespoke attribute to a class.

    This entry point will not be able to display any documentation but it is a shortcut to test new features.
    If the value is a Javascript object, the PyJs object must be used

    Attributes:
    ----------
    :param name: String. The key to be added to the attributes
    :param value: String or JString. The value of the defined attributes

    :return: The DataAttrs to allow the chains
    """
    self._attrs[name] = value
    return self

  def attr(self, name, value):
    """
    Description:
    ------------
    Add an attribute to the Javascript underlying dictionary

    Attributes:
    ----------
    :param name: String. The attribute name
    :param value: Object. The attribute value

    :return: "Self" to allow the chains on the Python side
    """
    self._attrs[name] = value
    return self

  def has_attribute(self, clsObj):
    """
    Description:
    ------------
    Add an extra sub layer to the data structure.
    The key in the object representation will be the function name

    Attributes:
    ----------
    :param clsObj: Class. The sub data class used in the structure definition
    """
    return self.sub_data(sys._getframe().f_back.f_code.co_name, clsObj)

  def get(self, dflt=None, name=None):
    """
    Description:
    ------------
    Get tje attribute to the underlying attributes dictionary

    Attributes:
    ----------
    :param dflt: Optional. The default value of this attribute
    :param name: Optional. The attribute name. default the name of the function

    :return:
    """
    return self._attrs.get(name or sys._getframe().f_back.f_code.co_name, dflt)

  def set(self, value, name=None):
    """
    Description:
    ------------
    Add an attribute to the Javascript underlying dictionary

    Attributes:
    ----------
    :param value: Object. The attribute value
    :param name: Optional. The attribute name. default the name of the function

    :return: "Self" to allow the chains on the Python side
    """
    return self.attr(name or sys._getframe().f_back.f_code.co_name, value)

  def sub_data(self, name, clsObj):
    """
    Description:
    ------------
    Add an extra sub layer to the data structure.
    The key in the object representation will be the function name

    Should use has_attribute is the name can be deduced from the parent function

    Attributes:
    ----------
    :param name: String. The key to be added to the internal data dictionary
    :param clsObj: Object. The object which will be added to the nexted data structure
    """
    if name in self._attrs:
      return self._attrs[name]

    self.__sub_levels.add(name)
    self._attrs[name] = clsObj(self._report)
    return self._attrs[name]

  def sub_data_enum(self, name, clsObj):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param name: String. The key to be added to the internal data dictionary
    :param clsObj: Object. The object which will be added to the nexted data structure
    """
    #if name in self._attrs:
    #  return self._attrs[name]

    self.__sub__enum_levels.add(name)
    enum_data = clsObj(self._report)
    self._attrs.setdefault(name, []).append(enum_data)
    return enum_data

  def __str__(self):
    result = []
    for s in self.__sub_levels:
      result.append("%s: %s" % (s, str(self._attrs[s])))
    for s in self.__sub__enum_levels:
      result.append("%s: [%s]" % (s, ",".join([str(k) for k in self._attrs[s]])))
    result.extend(["%s: %s" % (k, JsUtils.jsConvertData(v, None)) for k, v in self._attrs.items() if k not in self.__sub_levels and k not in self.__sub__enum_levels])
    return "{%s}" % ", ".join(result)


class DataGroup(object):

  def __init__(self, report, attrs, parent=None):
    self._attrs, self._report, self._parent = attrs, report, parent


class DataEnum(object):

  dflt = None
  js_conversion = False

  def __init__(self, report, value=None):
    self._report, self.__value = report, value or self.dflt

  def set(self, value=None):
    """
    Description:
    ------------
    Set the selected value in this enumeration.
    The last function call will be persisted

    Attributes:
    ----------
    :param value: Optional. The value to be set (default is the function name)
    """
    if value is None:
      value = sys._getframe().f_back.f_code.co_name
    if self.js_conversion:
      value = JsUtils.jsConvertData(value, None).toStr()
    self.__value = value

  def custom(self, value):
    """
    Description:
    ------------
    Set a custom value.This will not use any specific conversion

    Attributes:
    ----------
    :param value: String. The value to be set
    """
    self.__value = value

  def __str__(self):
    return self.__value


class DataEnumMulti(object):

  dflt = None
  js_conversion = False
  delimiter = ","

  def __init__(self, report, value=None):
    self._report = report
    value = value or self.dflt
    self.__value = set() if value is None else set([value])

  def set(self, value=None):
    """
    Description:
    ------------
    Set the selected value in this enumeration.
    The last function call will be persisted

    Attributes:
    ----------
    :param value: Optional. The value to be set (default is the function name)
    """
    if value is None:
      value = sys._getframe().f_back.f_code.co_name
    self.__value.add(value)

  def custom(self, value):
    """
    Description:
    ------------
    Set a custom value.This will not use any specific conversion

    Attributes:
    ----------
    :param value: String. The value to be set
    """
    self.__value.add(value)

  def __str__(self):
    result = self.delimiter.join(list(self.__value))
    if self.js_conversion:
      return JsUtils.jsConvertData(result, None).toStr()

    return result




import json

from epyk.core.html import Html
from epyk.core.html.options import OptSelect

# The list of CSS classes
from epyk.core.css.styles import GrpClsList

#
from epyk.core.js import JsUtils
from epyk.core.js.html import JsHtmlSelect
from epyk.core.js.packages import JsQuery
from epyk.core.js.packages import JsSelect


class Option(Html.Html):
  name, category, callFnc = 'Select Option', 'Lists', None
  builder_name = False

  def __init__(self, report, value, text, icon, selected):
    super(Option, self).__init__(report, text)
    self.set_attrs(name="value", value=value)
    if selected:
      self.set_attrs(name="selected", value=selected)
    if icon is not None:
      self.set_attrs(name="data-icon", value=icon)

  def __str__(self):
    return "<option %s>%s</option>" % (self.get_attrs(pyClassNames=self.style.get_classes()), self.val)


class Optgroup(Html.Html):
  name, category, callFnc = 'Select Option', 'Lists', None
  builder_name = False

  def __init__(self, report, data, label):
    super(Optgroup, self).__init__(report, data)
    self.set_attrs(name="label", value=label)

  def __str__(self):
    val = "".join([v.html() for v in self.val])
    return "<optgroup %s>%s</optgroup>" % (self.get_attrs(pyClassNames=self.style.get_classes()), val)


class Select(Html.Html):
  __reqCss, __reqJs = ['select'], ['select']
  name, category, callFnc = 'Select', 'Lists', 'select'

  def __init__(self, report, records, htmlCode, width, height, filter, profile, multiple, options):
    super(Select, self).__init__(report, records, htmlCode=htmlCode, css_attrs={"width": width, "height": height},
                                 globalFilter=filter, profile=profile)
    self.selected = None
    if multiple:
      self.attr['multiple'] = None
    self.__options = OptSelect.OptionsSelect(self, options)

  @property
  def options(self):
    """
    Property to set all the possible object for a button

    :rtype: OptSelect.OptionsSelect
    """
    return self.__options

  @property
  def style(self):
    """

    :rtype: GrpClsList.ClassSelect
    """
    if self._styleObj is None:
      self._styleObj = GrpClsList.ClassSelect(self)
    return self._styleObj

  @property
  def dom(self):
    """
    Description:
    -----------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsHtmlSelect.DomSelect
    """
    if self._dom is None:
      self._dom = JsHtmlSelect.DomSelect(self, report=self._report)
    return self._dom

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/methods/

    Attributes:
    ----------
    :return: A Javascript Dom object
    :rtype: JsSelect.JSelect
    """
    if self._js is None:
      self._js = JsSelect.JSelect(self, report=self._report)
    return self._js

  @property
  def _js__builder__(self):
    return '''
      var selectObj = %s; selectObj.empty(); 
      for (var idx in data) {
          const text = (typeof data[idx].text !== 'undefined')? data[idx].text : data[idx].value;
          selectObj.append('<option value=' + data[idx].value + '>' + text + '</option>'); }
      selectObj.selectpicker(options).val().selectpicker('refresh');''' % JsQuery.decorate_var("htmlObj", convert_var=False)

  def change(self, jsFncs, emtpyFncs=None, profile=False):
    """
    Description:
    -----------
    Javascript event triggered when the value has changed

    Attributes:
    ----------
    :param jsFncs: List. Set of Javascript function to trigger on this event
    :param emtpyFncs: ist. Set of Js function to trigger if the value is empty
    :param profile: Boolean. To set the profiling
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    if emtpyFncs is not None:
      if not isinstance(emtpyFncs, list):
        emtpyFncs = [emtpyFncs]
      jsFncs.append("if (%s === ''){ %s }" % (self.dom.content.toStr(), JsUtils.jsConvertFncs(emtpyFncs, toStr=True)))
    return self.on("change", jsFncs, profile)

  def __str__(self):
    options, opt_groups = [], {}
    for val in self.val:
      opt = Option(self._report, val['value'], val['name'], None, self.selected is not None and self.selected == val['value'])
      opt.inReport = False
      if 'group' in val:
        opt_groups.setdefault(val['group'], []).append(opt)
      else:
        options.append(opt.html())
    data = options
    for k in sorted(opt_groups):
      opt_rp = Optgroup(self._report, opt_groups[k], k)
      opt_rp.inReport = False
      data.append(opt_rp.html())
    self._report._props.setdefault('js', {}).setdefault("builders", []).append("%s.selectpicker(%s).selectpicker('refresh')" % (JsQuery.decorate_var(self.dom.varId, convert_var=False), json.dumps(self._jsStyles)))
    if self.attr.get("data-width") is not None:
      self._report.css.customText('.%s_width {width: %s !IMPORTANT}' % (self.htmlId, self.attr.get("data-width")))
    self.attr['class'].add("%s_width" % self.htmlId)
    data_cls = self.get_attrs(pyClassNames=self.style.get_classes()).replace('class="', 'data-style="')
    return "<select %s>%s</select>" % (data_cls, "".join(data))

  # -----------------------------------------------------------------------------------------
  #                                    EXPORT OPTIONS
  # -----------------------------------------------------------------------------------------
  def to_xls(self, workbook, worksheet, cursor):
    if self.htmlId in self._report.http:
      cell_title = self._jsStyles["title"] if self._jsStyles.get("title") is not None else 'Input'
      cell_format = workbook.add_format({'bold': True})
      worksheet.write(cursor['row'], 0, cell_title, cell_format)
      cursor['row'] += 1
      worksheet.write(cursor['row'], 0, self._report.http[self.htmlId])
      cursor['row'] += 2

  def to_word(self, document):
    p = document.add_paragraph()
    p.add_run("Selected: ")
    selected = ""
    for rec in self.vals:
      if rec.get('selected', False):
        selected = rec['value']
    runner = p.add_run(selected)
    runner.bold = True


class Lookup(Select):
  __reqCss, __reqJs = ['select'], ['select']

  def __init__(self, report, records, htmlCode, width, height, filter, profile, multiple, options):
    super(Lookup, self).__init__(report, records, htmlCode, width, height, filter, profile, multiple, options)
    self._jsStyles['lookups'] = records

  @property
  def _js__builder__(self):
    return '''
        var selectObj = %s; selectObj.empty(); const lookupData = options.lookups[data];
        if (typeof lookupData !== 'undefined') {
          for (var idx in lookupData) {
              const text = (typeof lookupData[idx].text !== 'undefined')? lookupData[idx].text : lookupData[idx].value;
              selectObj.append('<option value=' + lookupData[idx].value + '>' + text + '</option>'); }
          selectObj.selectpicker(options).val(lookupData).selectpicker('refresh')}
        else {selectObj.selectpicker(options).selectpicker('refresh')} ''' % JsQuery.decorate_var("htmlObj", convert_var=False)

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append("%s.selectpicker().selectpicker('refresh')" % JsQuery.decorate_var(self.dom.varId, convert_var=False))
    data_cls = self.get_attrs(pyClassNames=self.style.get_classes()).replace('class="', 'data-style="')
    return "<select %s></select>" % data_cls



"""

"""


class CatalogGroup(object):
  def __init__(self, report, class_list_type, html_id=None):
    self.__rptObj, self.__class_list_type, self._html_id = report, class_list_type, html_id

  def _set_class(self, classObj):
    cssObj = classObj(self.__rptObj, html_id=self._html_id)
    cssObj.customize()
    self.__class_list_type.add(cssObj)
    return cssObj



"""
Entry point for the different Javascript primitives

"""

import json

from epyk.core.js.primitives import JsArray
from epyk.core.js.primitives import JsDate
from epyk.core.js.primitives import JsObject
from epyk.core.js.primitives import JsNumber
from epyk.core.js.primitives import JsString
from epyk.core.js.primitives import JsBoolean

from epyk.core.js.objects import JsNodeDom
from epyk.core.js.objects import JsData
from epyk.core.js.objects import JsEvents

from epyk.core.js import JsUtils


class JsObjects(object):
  def __init__(self, jsObj):
    self._jsObj = jsObj

  @property
  def this(self):
    """
    Description:
    -----------
    Interface to the Javascript Object primitive

    :return: The Javascript "this" object (which is a dom object very often)
    """
    return JsNodeDom.JsDoms.get("this")

  @property
  def result(self):
    """
    Description:
    -----------
    Interface to the Javascript Object primitive

    :return: The Javascript result object from a post or any other AJAX calls
    """
    return JsObject.JsObject.get("result")

  @property
  def request(self):
    """
    Description:
    -----------
    Interface to the Javascript Object primitive

    :return: The Javascript result object from a post or any other AJAX calls

    :rtype: XMLHttpRequest
    """
    return XMLHttpRequest.get("result")

  @property
  def data(self):
    """
    Description:
    -----------
    Interface to the Javascript Object primitive

    :return: The Javascript data object from a post or any other AJAX calls
    """
    return JsObject.JsObject.get("data")

  @property
  def mouseEvent(self):
    """
    Description:
    -----------
    Interface to the JavaScript MouseEvents

    Related Pages:

			https://www.w3schools.com/jsref/obj_mouseevent.asp
    """
    return JsEvents.MouseEvent()

  @property
  def event(self):
    """
    Description:
    -----------
    Interface to the JavaScript Events

    Related Pages:

			https://www.w3schools.com/jsref/obj_event.asp
    """
    return JsEvents.Event()

  @property
  def jqThis(self):
    """
    Description:
    -----------
    Interface to the Javascript Object primitive

    :return: The Javascript "this" object
    """
    from epyk.core.js.packages import JsQuery
    return JsQuery.JQuery(self._jsObj._src, selector="jQuery(this)", setVar=False)

  @classmethod
  def get(cls, varName):
    """
    Description:
    -----------
    Interface to the Javascript Object primitive

    :return: The requested Python JsObject primitive
    """
    return JsObject.JsObject.get(varName)

  @classmethod
  def new(cls, data=None, varName=None, isPyData=False):
    """
    Description:
    -----------
    Interface to the Javascript Object primitive

    :return: A Python generic JsObject primitive
    """
    return JsObject.JsObject.new(data, varName, isPyData)

  @property
  def number(self):
    """
    Description:
    -----------
    Interface to the Javascript Number primitive

    Related Pages:

			https://www.w3schools.com/jsref/jsref_number.asp

    :return: A Python JsNumber primitive
    """
    return JsNumber.JsNumber

  @property
  def string(self):
    """
    Description:
    -----------
    Interface to the Javascript String primitive

    Related Pages:

			https://www.w3schools.com/jsref/jsref_obj_string.asp

    :return: A Python JsString primitive
    """
    return JsString.JsString

  @property
  def array(self):
    """
    Description:
    -----------
    Interface to the Javascript Array primitive

    Related Pages:

			https://www.w3schools.com/jsref/jsref_obj_array.asp

    :return: A Python JsArray primitive
    """
    return JsArray.JsArray

  @property
  def date(self):
    """
    Description:
    -----------
    Interface to the Javascript Date primitive

    Related Pages:

			https://www.w3schools.com/jsref/jsref_obj_date.asp

    :return: A Python JsDate object
    """
    return JsDate.JsDate

  @property
  def boolean(self):
    """
    Description:
    -----------
    Interface to the Javascript Boolean primitive

    Related Pages:

			https://www.w3schools.com/jsref/jsref_obj_boolean.asp

    :return: A Python JsBoolean primitive
    """
    return JsBoolean.JsBoolean

  @property
  def dom(self):
    """
    Description:
    -----------
    Interface to the Javascript Dom class

    :return: A Python Document
    """
    return JsNodeDom.JsDoms

  @property
  def null(self):
    """
    Description:
    -----------
    Similar as None in Python

    Related Pages:

			https://www.w3schools.com/js/js_datatypes.asp

    :return: A Python Js Null object
    """
    return JsObject.JsObject("null")

  @property
  def undefined(self):
    """
    Description:
    -----------
    Similar as the None in Python

    Related Pages:

			https://www.w3schools.com/jsref/jsref_undefined.asp

    :return: A Python Js undefined object
    """
    return JsObject.JsObject("undefined", isPyData=False)

  @property
  def NaN(self):
    """
    Description:
    -----------
    The NaN property represents "Not-a-Number" value. This property indicates that a value is not a legal number.

    The NaN property is the same as the Number.Nan property.

    Related Pages:

:
    https://www.w3schools.com/jsref/jsref_number_nan.asp
    https://www.w3schools.com/jsref/jsref_isnan.asp
    """
    return JsNumber.JsNumber("Number.NaN")

  @property
  def true(self):
    """
    Description:
    -----------
    Similar as True in Python

    Related Pages:

			https://www.w3schools.com/js/js_booleans.asp

    :return: A Python Js True object
    """
    return JsBoolean.JsBoolean.get('true')

  @property
  def false(self):
    """
    Description:
    -----------
    Similar as False in Python

    Related Pages:

			https://www.w3schools.com/js/js_booleans.asp
    
    :return: A Python Js False object
    """
    return JsBoolean.JsBoolean.get('false')

  def record(self, varName):
    """
    Description:
    -----------
    Get a record object

    :param varName: A string with of the existing variable name
    """
    return JsData.RawData.get(self._jsObj, varName)


class JsPromise(object):

  def __init__(self, jsObj):
    self._jsObj = jsObj
    self.__then, self.__catch = [], []

  def then(self, jsFnc):
    """
    Description:
    -----------

    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = []
    self.__then.extend(jsFnc)
    return self

  def catch(self, jsFnc):
    """
    Description:
    -----------

    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = []
    self.__catch.extend(jsFnc)
    return self

  @property
  def r(self):
    return self.toStr()

  def toStr(self):
    result = [str(self._jsObj)]
    if self.__then:
      result.append("then(function(fulfilled){%s})" % ";".join(self.__then))
    if self.__catch:
      result.append("catch(function(error){%s})" % ";".join(self.__catch))
    return ".".join(result)

  def __str__(self):
    return self.toStr()


class XMLHttpRequest(object):

  def __init__(self, report, varName, method_type, url, data=None):
    self.data = {} if data is None else data
    self._src, self.__headers, self.url = report, {}, url
    self.__mod_name, self.__mod_path, self.method = None, None, method_type
    self.__req_success, self.__req_fail, self.__req_send = None, None, None
    self.__url_prefix = ""
    self.varId = varName
    if url is not None:
      self.open(method_type, url)

  @classmethod
  def get(cls, varName):
    """
    Description:
    ------------
    Interface to the Javascript Object primitive

    :return: The requested Python JsObject primitive
    """
    return XMLHttpRequest(None, varName, None, None)

  def readyState(self):
    """
    Description:
    ------------
    The XMLHttpRequest.readyState property returns the state an XMLHttpRequest client is in.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/readyState
    """
    return JsNumber.JsNumber("%s.readyState" % self.varId)

  def status(self):
    """
    Description:
    ------------
    The read-only XMLHttpRequest.status property returns the numerical HTTP status code of the XMLHttpRequest's response.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/status
    """
    return JsNumber.JsNumber("%s.status" % self.varId)

  def responseType(self):
    """
    Description:
    ------------
    The XMLHttpRequest property responseType is an enumerated string value specifying the type of data contained in the response.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/responseType
    """
    return JsString.JsString("%s.responseType" % self.varId, isPyData=False)

  @property
  def responseText(self):
    """
    Description:
    ------------
    The read-only XMLHttpRequest property responseText returns the text received from a server following a request being sent.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/responseText
    """
    return JsString.JsString("%s.responseText" % self.varId, isPyData=False)

  def timeout(self, milliSeconds):
    """
    Description:
    ------------
    The XMLHttpRequest.timeout property is an unsigned long representing the number of milliseconds a request can take before automatically being terminated.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/timeout

    Attributes:
    ----------
    :param milliSeconds: Number. The value of milliseconds
    """

  def abort(self):
    """
    Description:
    ------------
    The abort event is fired when a request has been aborted, for example because the program called XMLHttpRequest.abort().

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/abort_event
    """

  def addEventListener(self, event, jsFncs):
    """
    Description:
    ------------

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/error_event

    Attributes:
    ----------
    :param event:
    :param jsFncs:
    """

  def open(self, method_type, url, _async=True, user=None, password=None):
    """
    Description:
    ------------
    The XMLHttpRequest method open() initializes a newly-created request, or re-initializes an existing one.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/open

    Attributes:
    ----------
    :param method_type: String. The HTTP request method to use, such as "GET", "POST", "PUT", "DELETE", etc. Ignored for non-HTTP(S) URLs.
    :param url: String. A DOMString representing the URL to send the request to.
    :param _async: Boolean. An optional Boolean parameter, defaulting to true, indicating whether or not to perform the operation asynchronously.
    :param user: The optional user name to use for authentication purposes; by default, this is the null value.
    :param password: The optional password to use for authentication purposes; by default, this is the null value.
    """
    self.url, self.method = url, method_type
    return self

  def setHeaders(self, headers):
    """
    Description:
    ------------
    The XMLHttpRequest method setRequestHeader() sets the value of an HTTP request header.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/setRequestHeader

    Attributes:
    ----------
    :param headers: Dictionary. The different attributes to be added to the header
    """
    self.__headers.update(headers)
    return self

  def setRequestHeader(self, name, value):
    """
    Description:
    ------------
    The XMLHttpRequest method setRequestHeader() sets the value of an HTTP request header.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/setRequestHeader

    Attributes:
    ----------
    :param name: String. The header name
    :param value: String. The header value
    """
    self.__headers[name] = value
    return self

  def onSuccess(self, jsFncs):
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self.__req_success = jsFncs
    return self

  def ontimeout(self, jsFncs):
    """
    Description:
    ------------

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/timeout

    Attributes:
    ----------
    :param jsFncs:
    """

  def onFail(self, jsFncs):
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self.__req_fail = jsFncs
    return self

  def withCredentials(self, flag):
    """
    Description:
    ------------
    The XMLHttpRequest.withCredentials property is a Boolean that indicates whether or not cross-site Access-Control requests should be made using credentials such as cookies, authorization headers or TLS client certificates. Setting withCredentials has no effect on same-site requests.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/withCredentials

    Attributes:
    ----------
    :param flag:
    """

  def send(self, jsonData=None, encodeURIData=None):
    """
    Description:
    ------------
    The XMLHttpRequest method send() sends the request to the server.
    If the request is asynchronous (which is the default), this method returns as soon as the request is sent and the result is delivered using events.
    If the request is synchronous, this method doesn't return until the response has arrived.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/send

    Attributes:
    ----------
    :param jsonData:
    :param encodeURIData:
    """
    #Initialize jsonData with potential initial data passed in the constructor
    if jsonData:
      jsonData.update(self.data)
    else:
      jsonData = self.data
    if jsonData:
      self.__req_send = "%s.send(JSON.stringify(%s))" % (self.varId, json.dumps(jsonData))
    elif encodeURIData is not None:
      self.__url_prefix = "?%s" % "&".join(["%s=%s" % (k, v) for k, v in encodeURIData.items()])
      self.__req_send = "%s.send()" % self.varId
    else:
      self.__req_send = "%s.send()" % self.varId
    return self

  def toStr(self):
    request = ["var %s = new XMLHttpRequest()" % self.varId]
    request.append("%s.open(%s, '%s%s')" % (self.varId, self.method, self.url, self.__url_prefix))
    for k, v in self.__headers.items():
      request.append("%s.setRequestHeader('%s', '%s')" % (self.varId, k, v))
    if self.__req_success is not None:
      if self.__req_fail is not None:
        request.append("%s.onload = function() {}" % self.varId)
      else:
        request.append("%s.onload = function() {%s}" % (self.varId, JsUtils.jsConvertFncs(self.__req_success, toStr=True)))
    if self.__req_send is None:
      raise Exception("The send method must be called")

    request.append(self.__req_send)
    return ";".join(request)



"""
CSS Style module for the Table components
"""

from epyk.core.css.styles.classes import CssStyle


class CssDataTable(CssStyle.Style):
  _attrs = {'border-collapse': 'collapse !IMPORTANT'}

  def customize(self):
    self.css({'border': "1px solid %s" % self.rptObj.theme.greys[3]}, important=True)
    self.hover.css({'border': "1px solid %s" % self.rptObj.theme.success[1]}, important=True)


class CssDataTableHeader(CssStyle.Style):
  _selectors = {'child': 'thead'}

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[0], "background": self.rptObj.theme.greys[-1], 'white-space': 'nowrap'})


class CssDataTableFooter(CssStyle.Style):
  _selector = 'paginate_button'
  classname = False

  def customize(self):
    self.css({"color": "red", 'background-color': "%s" % self.rptObj.theme.colors[2]}, important=True)


class CssDataTableEven(CssStyle.Style):
  _selectors = {'child': 'tbody tr:nth-child(even)'}

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'background-color': self.rptObj.theme.greys[0],
              'border-top': "1px solid %s !IMPORTANT" % self.rptObj.theme.greys[0]})
    self.hover.css({'border': "1px solid %s !IMPORTANT" % self.rptObj.theme.success[1],
                    'color': self.rptObj.theme.greys[-1], 'background-color': self.rptObj.theme.colors[3]})


class CssDataTableOdd(CssStyle.Style):
  _selectors = {'child': 'tbody tr:nth-child(odd)'}

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'background-color': self.rptObj.theme.colors[1],
              'border-top': "1px solid %s !IMPORTANT" % self.rptObj.theme.colors[1]})
    self.hover.css({'border': "1px solid %s !IMPORTANT" % self.rptObj.theme.success[1],
                    'color': self.rptObj.theme.greys[-1], 'background-color': self.rptObj.theme.colors[3]})


class CssTableBasic(CssStyle.Style):
  _attrs = {'margin': '5px', 'border-collapse': 'collapse'}


# class CssTableColumnFixed(CssStyle.Style):
#   _attrs = {'margin': '5px', 'text-align': 'left', 'font-weight': 'bold'}
#

class CssTableNewRow(CssStyle.Style):
  _attrs = {'color': '#546472'}


class CssTableSelected(CssStyle.Style):
  _attrs = {'background-color': '#AEDAF8 !important'}


class CssCellComment(CssStyle.Style):
  _attrs = {'margin': '0!important', 'padding': '2px 0 0 2px!important'}


class CssCellSave(CssStyle.Style):
  _attrs = {'color': '#293846!important'}


class CssTdEditor(CssStyle.Style):
  _attrs = {'border-width': '1px', 'border-style': 'solid', 'text-align': 'left', 'height': '30px', 'padding': '5px',
            'vertical-align': 'middle'}

  def customize(self):
    self.css({"color": self.rptObj.theme.colors[5], 'border-color': self.rptObj.theme.greys[5]})


class CssTdDetails(CssStyle.Style):
  _before = {'content': "'\\f0fe'", 'font-family': "'Font Awesome 5 Free'", 'cursor': 'pointer', 'padding': '0 5px 0 0'}
  _selectors = {'child': 'td'}


class CssTdDetailsShown(CssStyle.Style):
  _before = {'content': "'\\f146'", 'font-family': "'Font Awesome 5 Free'", 'cursor': 'pointer', 'padding': '0 5px 0 0'}
  _selectors = {'child': 'td'}


class CssTdGridHeaderCols(CssStyle.Style):
  _attrs = {"border-bottom": '1px solid black', "margin": '0 1px'}
  _selectors = {'child': 'th:not(:first-child)'}


class CssTdGridNoHeaderCols(CssStyle.Style):
  _attrs = {"display": 'none'}
  _selectors = {'child': 'th:not(:first-child)'}


class CssTdGridHeaderRows(CssStyle.Style):
  _attrs = {"text-align": 'left'}
  _selectors = {'child': 'td[name="row_header"]'}


class CssTdGridVals(CssStyle.Style):
  _selectors = {'child': 'td:not([name="row_header"])'}
  _attrs = {"padding": '3px'}
  _focus = {"outline": "0px solid transparent"}

  def customize(self):
    self.css({"background-color": self.rptObj.theme.colors[0],
              "border-bottom": '1px solid %s' % self.rptObj.theme.greys[4]})
    self.focus.css({"border-bottom": '1px solid %s' % self.rptObj.theme.success[1]})




from epyk.core.js.packages import packageImport
from epyk.core.html.options import Options
from epyk.core.js import Imports


class OptionsCode(Options):

  @property
  def value(self):
    """
    Description:
    ------------
    The starting value of the editor. Can be a string, or a document object.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(True)

  @value.setter
  def value(self, value):
    self._config(value)

  @property
  def mode(self):
    """
    Description:
    ------------
    The mode to use. When not given, this will default to the first mode that was loaded.
    It may be a string, which either simply names the mode or is a MIME type associated with the mode.

    add extra addon: /mode/javascript/javascript.js

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get("")

  @mode.setter
  def mode(self, value):
    Imports.extend('codemirror-%s' % value, [('%s.min.js' % value, 'codemirror/%%(version)s/mode/%s/' % value)], version="codemirror")
    self._report.jsImports.add('codemirror-%s' % value)
    self._config(value)

  @property
  def lineSeparator(self):
    """
    Description:
    ------------
    Explicitly set the line separator for the editor.
    By default (value null), the document will be split on CRLFs as well as lone CRs and LFs, and a single LF will be used as line separator in all output (such as getValue).
    When a specific string is given, lines will only be split on that string, and output will, by default, use that same separator.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get()

  @lineSeparator.setter
  def lineSeparator(self, value):
    self._config(value)

  @property
  def theme(self):
    """
    Description:
    ------------
    The theme to style the editor with.
    You must make sure the CSS file defining the corresponding .cm-s-[name] styles is loaded (see the theme directory in the distribution).
    The default is "default", for which colors are included in codemirror.css.
    It is possible to use multiple theming classes at once—for example "foo bar" will assign both the cm-s-foo and the cm-s-bar classes to the editor.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get()

  @theme.setter
  def theme(self, value):
    self._config(value)

  @property
  def indentUnit(self):
    """
    Description:
    ------------
    How many spaces a block (whatever that means in the edited language) should be indented. The default is 2.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(2)

  @indentUnit.setter
  def indentUnit(self, value):
    self._config(value)

  @property
  def smartIndent(self):
    """
    Description:
    ------------
    Whether to use the context-sensitive indentation that the mode provides (or just indent the same as the line before). Defaults to true.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(True)

  @smartIndent.setter
  def smartIndent(self, value):
    self._config(value)

  @property
  def tabSize(self):
    """
    Description:
    ------------
    The width of a tab character. Defaults to 4.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(4)

  @tabSize.setter
  def tabSize(self, value):
    self._config(value)

  @property
  def indentWithTabs(self):
    """
    Description:
    ------------
    Whether, when indenting, the first N*tabSize spaces should be replaced by N tabs. Default is false.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(False)

  @indentWithTabs.setter
  def indentWithTabs(self, value):
    self._config(value)

  @property
  def electricChars(self):
    """
    Description:
    ------------
    Configures whether the editor should re-indent the current line when a character is typed that might change its proper indentation (only works if the mode supports indentation).
    Default is true.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(True)

  @electricChars.setter
  def electricChars(self, value):
    self._config(value)

  @property
  def direction(self):
    """
    Description:
    ------------
    Flips overall layout and selects base paragraph direction to be left-to-right or right-to-left.
    Default is "ltr". CodeMirror applies the Unicode Bidirectional Algorithm to each line, but does not autodetect base direction — it's set to the editor direction for all lines.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get("ltr")

  @direction.setter
  def direction(self, value):
    self._config(value)

  @property
  def rtlMoveVisually(self):
    """
    Description:
    ------------
    Determines whether horizontal cursor movement through right-to-left (Arabic, Hebrew) text is visual (pressing the left arrow moves the cursor left) or logical (pressing the left arrow moves to the next lower index in the string, which is visually right in right-to-left text). The default is false on Windows, and true on other platforms.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(False)

  @rtlMoveVisually.setter
  def rtlMoveVisually(self, value):
    self._config(value)

  @property
  def keyMap(self):
    """
    Description:
    ------------
    Configures the key map to use. The default is "default", which is the only key map defined in codemirror.js itself.
    Extra key maps are found in the key map directory. See the section on key maps for more information.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get("default")

  @keyMap.setter
  def keyMap(self, value):
    self._config(value)

  @property
  def lineWrapping(self):
    """
    Description:
    ------------
    Whether CodeMirror should scroll or wrap for long lines. Defaults to false (scroll).

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(False)

  @lineWrapping.setter
  def lineWrapping(self, value):
    self._config(value)

  @property
  def lineNumbers(self):
    """
    Description:
    ------------
    Whether to show line numbers to the left of the editor.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(True)

  @lineNumbers.setter
  def lineNumbers(self, value):
    self._config(value)

  @property
  def firstLineNumber(self):
    """
    Description:
    ------------
    At which number to start counting lines. Default is 1.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(1)

  @firstLineNumber.setter
  def firstLineNumber(self, value):
    self._config(value)

  @property
  def fixedGutter(self):
    """
    Description:
    ------------
    Determines whether the gutter scrolls along with the content horizontally (false) or whether it stays fixed during horizontal scrolling (true, the default).

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(True)

  @fixedGutter.setter
  def fixedGutter(self, value):
    self._config(value)

  @property
  def scrollbarStyle(self):
    """
    Description:
    ------------
    Chooses a scrollbar implementation. The default is "native", showing native scrollbars.
    The core library also provides the "null" style, which completely hides the scrollbars. Addons can implement additional scrollbar models.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get("native")

  @scrollbarStyle.setter
  def scrollbarStyle(self, value):
    self._config(value)

  @property
  def coverGutterNextToScrollbar(self):
    """
    Description:
    ------------
    When fixedGutter is on, and there is a horizontal scrollbar, by default the gutter will be visible to the left of this scrollbar.
    If this option is set to true, it will be covered by an element with class CodeMirror-gutter-filler.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(False)

  @coverGutterNextToScrollbar.setter
  def coverGutterNextToScrollbar(self, value):
    self._config(value)

  @property
  def inputStyle(self):
    """
    Description:
    ------------
    Selects the way CodeMirror handles input and focus.
    he core library defines the "textarea" and "contenteditable" input models. On mobile browsers, the default is "contenteditable". On desktop browsers, the default is "textarea".

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get("textarea")

  @inputStyle.setter
  def inputStyle(self, value):
    self._config(value)

  @property
  def readOnly(self):
    """
    Description:
    ------------
    This disables editing of the editor content by the user. If the special value "nocursor" is given (instead of simply true), focusing of the editor is also disallowed.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(False)

  @readOnly.setter
  def readOnly(self, value):
    self._config(value)

  @property
  def screenReaderLabel(self):
    """
    Description:
    ------------
    This label is read by the screenreaders when CodeMirror text area is focused. This is helpful for accessibility.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get("")

  @screenReaderLabel.setter
  def screenReaderLabel(self, value):
    self._config(value)

  @property
  def showCursorWhenSelecting(self):
    """
    Description:
    ------------
    Whether the cursor should be drawn when a selection is active. Defaults to false.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(False)

  @showCursorWhenSelecting.setter
  def showCursorWhenSelecting(self, value):
    self._config(value)

  @property
  def lineWiseCopyCut(self):
    """
    Description:
    ------------
    When enabled, which is the default, doing copy or cut when there is no selection will copy or cut the whole lines that have cursors on them.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(True)

  @lineWiseCopyCut.setter
  def lineWiseCopyCut(self, value):
    self._config(value)

  @property
  def pasteLinesPerSelection(self):
    """
    Description:
    ------------
    When pasting something from an external source (not from the editor itself), if the number of lines matches the number of selection,
    CodeMirror will by default insert one line per selection. You can set this to false to disable that behavior.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(False)

  @pasteLinesPerSelection.setter
  def pasteLinesPerSelection(self, value):
    self._config(value)

  @property
  def selectionsMayTouch(self):
    """
    Description:
    ------------
    Determines whether multiple selections are joined as soon as they touch (the default) or only when they overlap (true).

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(True)

  @selectionsMayTouch.setter
  def selectionsMayTouch(self, value):
    self._config(value)

  @property
  def historyEventDelay(self):
    """
    Description:
    ------------
    The period of inactivity (in milliseconds) that will cause a new history event to be started when typing or deleting. Defaults to 1250.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(1250)

  @historyEventDelay.setter
  def historyEventDelay(self, value):
    self._config(value)

  @property
  def tabindex(self):
    """
    Description:
    ------------
    The tab index to assign to the editor. If not given, no tab index will be assigned.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get()

  @tabindex.setter
  def tabindex(self, value):
    self._config(value)

  @property
  def autofocus(self):
    """
    Description:
    ------------
    Can be used to make CodeMirror focus itself on initialization. Defaults to off.
    When fromTextArea is used, and no explicit value is given for this option, it will be set to true when either the source textarea is focused, or it has an autofocus attribute and no other element is focused.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(False)

  @autofocus.setter
  def autofocus(self, value):
    self._config(value)

  @property
  def dragDrop(self):
    """
    Description:
    ------------
    Controls whether drag-and-drop is enabled. On by default.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(True)

  @dragDrop.setter
  def dragDrop(self, value):
    self._config(value)

  @property
  def cursorBlinkRate(self):
    """
    Description:
    ------------
    Half-period in milliseconds used for cursor blinking.
    The default blink rate is 530ms. By setting this to zero, blinking can be disabled. A negative value hides the cursor entirely.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(530)

  @cursorBlinkRate.setter
  def cursorBlinkRate(self, value):
    self._config(value)

  @property
  def cursorScrollMargin(self):
    """
    Description:
    ------------
    How much extra space to always keep above and below the cursor when approaching the top or bottom of the visible view in a scrollable document. Default is 0.

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(0)

  @cursorScrollMargin.setter
  def cursorScrollMargin(self, value):
    self._config(value)

  @property
  def cursorHeight(self):
    """
    Description:
    ------------
    Determines the height of the cursor. Default is 1, meaning it spans the whole height of the line.
    For some fonts (and by some tastes) a smaller height (for example 0.85), which causes the cursor to not reach all the way to the bottom of the line, looks better

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(1)

  @cursorHeight.setter
  def cursorHeight(self, value):
    self._config(value)

  @property
  def matchBrackets(self):
    """
    Description:
    ------------
    Defines an option matchBrackets which, when set to true or an options object, causes matching brackets to be highlighted whenever the cursor is next to them.
    It also adds a method matchBrackets that forces this to happen once, and a method findMatchingBracket that can be used to run the bracket-finding algorithm that this uses internally.
    It takes a start position and an optional config object.
    By default, it will find the match to a matchable character either before or after the cursor (preferring the one before), but you can control its behavior with these options:

    Add extra addon: addon/selection/active-line.js

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(False)

  @matchBrackets.setter
  def matchBrackets(self, value): # matchbrackets.js
    Imports.extend('codemirror-matchbrackets', [('matchbrackets.min.js', 'codemirror/%(version)s/addon/edit/')], version="codemirror")
    self._report.jsImports.add('codemirror-matchbrackets')
    self._config(value)

  @property
  def styleActiveLine(self):
    """
    Description:
    ------------
    Defines a styleActiveLine option that, when enabled, gives the wrapper of the line that contains the cursor the class CodeMirror-activeline, adds a background with the class CodeMirror-activeline-background, and adds the class CodeMirror-activeline-gutter to the line's gutter space is enabled.

    Add extra addon: addon/edit/matchbrackets.js

    Related Pages:

			https://codemirror.net/doc/manual.html#config
    """
    return self._config_get(False)

  @styleActiveLine.setter
  def styleActiveLine(self, value):
    Imports.extend('codemirror-active-line', [('active-line.min.js', 'codemirror/%(version)s/addon/selection/')], version="codemirror")
    self._report.jsImports.add('codemirror-active-line')
    self._config(value)
